{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeFamilies     #-}
{-# LANGUAGE QuasiQuotes      #-}
{-# OPTIONS_GHC -Wno-orphans  #-}

module Test.API.GraphQL (
    tests
  ) where

import Data.Aeson ((.=))
import Data.Aeson qualified as JSON
import Data.Vector qualified as V
import Gargantext.Core.Types.Individu
import Gargantext.Database.Admin.Types.Node (UserId(..))
import Gargantext.Prelude
import Servant.Auth.Client ()
import Test.API.Private (withValidLogin, protected, protectedNewError)
import Test.API.Setup (withTestDBAndPort)
import Test.Database.Setup (MasterUserEnv(..), getMasterUserEnvOrFail, getUserOrFail)
import Test.Hspec
import Test.Hspec.Wai.Internal (withApplication)
import Test.Hspec.Wai.JSON (json)
import Test.Utils
import Text.RawString.QQ (r)

tests :: Spec
tests = sequential $ aroundAll withTestDBAndPort $ do
  describe "GraphQL" $ do


    describe "get_user_infos" $ do
      it "allows 'alice' to see her own info" $ \((testEnv, port), app) -> do
        (aliceEmail, alicePassword, aliceId, _aliceNodeId) <- getUserOrFail testEnv "alice"
        
        withApplication app $ do
          withValidLogin port "alice" alicePassword $ \token -> do
            let query' = "{ user_infos(user_id:" <> show (_UserId aliceId) <> ") { ui_id, ui_email } }" :: Text
            let query = JSON.encode $ JSON.object [ "query" .= query' ]
            liftIO $ printDebug "[get_user_infos] query" query
            let userInfo = JSON.object [ "ui_id" .= aliceId
                                       ,  "ui_email" .= aliceEmail ]
            let expected = 
                  JSON.object [ "data" .= JSON.object [ "user_infos" .= JSON.Array (V.singleton userInfo) ] ]
            protected token "POST" "/gql" query `matchSimpleJSON` expected


    describe "check error format" $ do
      it "returns the new error if header X-Garg-Error-Scheme: new is passed" $ \((testEnv, port), app) -> do
        withApplication app $ do
          mue <- getMasterUserEnvOrFail testEnv
          
          withValidLogin port (userName mue) (GargPassword $ secretKey mue) $ \token -> do
            let query' = "{ languages(id:5) { lt_lang } }" :: Text
            let query = JSON.encode $ JSON.object [ "query" .= query' ]
            let location = JSON.object [ "column" .= (13 :: Int)
                                       , "line" .= (1 :: Int) ]
            let err = JSON.object [ "locations" .= JSON.Array (V.singleton location)
                                  , "message" .= ("Unknown Argument \"id\" on Field \"languages\"." :: Text) ]
            let expected = JSON.object [ "errors" .= JSON.Array (V.singleton err) ]
            protectedNewError token "POST" "/gql" query `matchSimpleJSON` expected

      it "returns the old error (though this is deprecated)" $ \((testEnv, port), app) -> do
        withApplication app $ do
          mue <- getMasterUserEnvOrFail testEnv
          
          withValidLogin port (userName mue) (GargPassword $ secretKey mue) $ \token -> do
            let query' = "{ languages(id:5) { lt_lang } }" :: Text
            let query = JSON.encode $ JSON.object [ "query" .= query' ]
            let location = JSON.object [ "column" .= (13 :: Int)
                                       , "line" .= (1 :: Int) ]
            let err = JSON.object [ "locations" .= JSON.Array (V.singleton location)
                                  , "message" .= ("Unknown Argument \"id\" on Field \"languages\"." :: Text) ]
            let expected = JSON.object [ "errors" .= JSON.Array (V.singleton err) ]
            protected token "POST" "/gql" query `matchSimpleJSON` expected

      it "check new errors with 'type'" $ \((testEnv, port), app) -> do
        withApplication app $ do
          mue <- getMasterUserEnvOrFail testEnv

          withValidLogin port (userName mue) (GargPassword $ secretKey mue) $ \token -> do
            let query = [r| { "query": "mutation { delete_team_membership(shared_folder_id:1, team_node_id:1, token:\"abc\") }" } |]
            let expected = [json| {"errors":[{"extensions":{"data":{"msg":"This user is not team owner","user_id":1},"diagnostic":"User not authorized. ","type":"EC_403__user_not_authorized"},"message":"User not authorized. "}]} |]
            shouldRespondWithFragmentCustomStatus 403
              (protectedNewError token "POST" "/gql" query)
              expected
